!******************************************************************************
!
! File : warp.f90
!   by : Alan Tackett based on an idea from Dr. Matthews
!   on : 12/28/00
!  for : Warping wavefunctions after atomic moves
!
! This module contains routines to warp the wave functions after the atoms
! have been moved.  Allowing much larger time steps.....
!
!
!******************************************************************************

Module warp

Use gpoints
Use psilib
Use paw_inout
Use misc
Use atom_data
Use crystal_data

Implicit NONE!!!!!!!!!!!!!!!!!!!!


Integer, PRIVATE, PARAMETER :: OLD_POS = 1  !** Old atom positions
Integer, PRIVATE, PARAMETER :: NEW_POS = 2  !** New Atom positions

Integer, PRIVATE, PARAMETER :: WARP_GRID = G_LOW  !** FFT grid to use


Type Warp_Context               !** Context containing transformation info
   Real, Pointer :: Pos(:,:,:)  !* Atom positions 
   Integer, Pointer :: Map(:)   !* Mapping between old grid and new grid
End type

!******************************************************************************
Contains
!******************************************************************************

!******************************************************************************
!
! CalcRealSpace - Calculates the Space point given the mesh coords
!
!    GridDim - Grid Dimensions
!    P       - Curvilinear Mesh Coords (INPUT)
!    R       - Real Space Coords  (OUTPUT)
!
!******************************************************************************

Subroutine CalcRealSpace(GridDim, P, R)
   Integer, Intent(IN)  :: GridDim(:)
   Integer, Intent(IN)  :: P(:)
   Real,   Intent(OUT) :: R(:)


   R = MATMUL(xtal%Basis, (P-1.0)/GridDim(1:3))

   Return
End Subroutine

!******************************************************************************
! CalcGridSpace - Calculates the mesh coords given the real space coords
!
! GridDim - Grid dimenstions
! R       - Real Space Coords  (INPUT)
! P       - Mesh corrds (OUTPUT)
!
!  NOTE: This routine does not check to make sure the calculated mesh point
!        lies within the grid!!
!
!******************************************************************************

Subroutine CalcGridSpace(GridDim, R, P)
   Integer,   Intent(IN)  :: GridDim(:)
   Real,      Intent(IN)  :: R(:)
   Real,      Intent(OUT) :: P(:)

   P = MATMUL(xtal%InverseBasis,R)*GridDim(1:3) + 1

   Return
End Subroutine

!******************************************************************************
! CalcWarpSpace - Calculates the warped coordinate given the real space coords
!
! Warp - Warp context containing the transformation
! R       - Real Space Coords  (INPUT)
! RW      - warped coords (OUTPUT)
!
!******************************************************************************

Subroutine CalcWarpSpace(Warp, R, RW)
  Type (Warp_context), Pointer :: Warp
   Real,           Intent(IN)  :: R(:)
   Real,           Intent(OUT) :: RW(:)

   !######### QWERT  Need to DO #####################

   Return
End Subroutine


!******************************************************************************
!
!  CalcGridOffset - Calculates the Offset into the Metric buffer containing
!                     the information for the specified mesh-point.
!
!  GridDim - Grid Dimensions
!  PM    - Mesh point 
!
!******************************************************************************

Function CalcGridOffset(GridDim, PM)
  Integer             :: CalcGridOffset
  Integer, Intent(IN) :: gridDim(:)
  Integer, Intent(IN) :: PM(:)

  CalcgridOffset = GridDim(1)*GridDim(2)*(PM(3)-1) + &
                     GridDim(1)*(PM(2)-1) + PM(1)-1

  Return
End Function


!******************************************************************************
!
!  WrapIndex - Returns the Grid point specified using periodic boundary
!      conditions.
!
!  GridDim - Grid Dim's
!  PM    - On Input it the mesh point to wrap and on exit it's wrapped.
!
!******************************************************************************

Subroutine WrapIndex(GridDim, PM)
   Integer, Intent(IN)    :: GridDim(:)
   Integer, Intent(INOUT) :: PM(:)

   Integer :: i, DX(3)
   Logical :: Ok

   Ok = .FALSE.
   DX = 0
   Do While (.NOT. Ok)
      Where (PM > GridDim(1:3))
         PM = PM - GridDim(1:3)
         DX = DX + 1
      End Where
         
      Where (PM < 1)
         PM = PM + GridDim(1:3)
         DX = DX - 1
      End Where

      Ok = .TRUE.
      If ((PM(1)<1) .OR. (PM(2)<1) .OR. (PM(3)<1)) Ok = .FALSE.

      If (Ok) then
         Do i=1, 3
            If (PM(i)>GridDim(i)) Ok = .FALSE.
         End Do
      End If
   End Do

   Return
End Subroutine


!******************************************************************************
!
!  Warp_Store_Pos - Stores the current atomic positions in the given slot
!
!       Warp - Warp Context where to store the data
!       Index - where to store the atomic positons(OLD_POS or NEW_POS)
!
!******************************************************************************

Subroutine Warp_Store_Pos(Warp, index)
  Type (Warp_context), Pointer :: Warp
  Integer,          Intent(IN) :: Index

  Integer :: i
  real, Pointer :: Pos(:,:)

  Pos => Warp%Pos(:,:,Index)  !** Assign pointer 

  Do i=1, Specific_Atoms
     Pos(:,i) = Atom_list(i)%Pos
  End Do

  Return
End Subroutine


!******************************************************************************
!
! Create_Warp - Creates a NEW warp context
!
!******************************************************************************

Subroutine Create_Warp(Warp)
  Type (Warp_Context), Pointer :: Warp
  
  Integer :: ierr

  !** These are small arrays so shouldn't be a problem allocating them **
  Allocate(Warp)
  Allocate(Warp%Pos(3,2,Specific_Atoms))

  !** But this is larger so we check **
  Allocate(Warp%Map(FFT_Grid(4,WARP_GRID)), STAT=ierr)
  Call Check_Error(ierr, "Could Not Allocate Map!", Error_Unit, &
       .TRUE.,paw_wc, "Create_Warp:")
 
  Return
End Subroutine


!******************************************************************************
!
! Destroy_Warp - Destroys a Warp context
!
!******************************************************************************

Subroutine Destroy_Warp(Warp)
  Type (Warp_Context), Pointer :: Warp
  

  DeAllocate(Warp%Map, Warp%Pos)  !** Delete the data structure elements

  DeAllocate(Warp)   !** Lastly delete the context itself

  Return
End Subroutine

!******************************************************************************
!
!  Warp_BeforeMove - Routine to Call before Atom's are moved.  This
!       routine simply stores the current atom positions
!
!       Warp - Warp context for storing information in
!
!******************************************************************************

Subroutine Warp_BeforeMove(Warp)
  Type (Warp_context), Pointer :: Warp

  Call Create_Warp(Warp)               !** Create the context
  Call Warp_Store_Pos(Warp, OLD_POS)   !** Copy the atomic positions

  Return
End Subroutine

!******************************************************************************
!
!  Warp_create_Transform - Creates the transform from the old/new pos
!
!           Warp - Context containing the positions and also where to
!                  store the transformation
!
!******************************************************************************

Subroutine Warp_Create_Transform(Warp)
  Type (Warp_context), Pointer :: Warp

  !############# QWERT Need to fix this !!!!! #####################

  Return
End subroutine

!******************************************************************************
!
!  Warp_Create_Mapping - Creates the linear mapping between grids
!
!         Warp - Context to use
!
!******************************************************************************

Subroutine Warp_Create_Mapping(Warp)
  Type (Warp_context), Pointer :: Warp

  Integer :: x,y,z, i, j, PM(3), grid(3)
  Real    :: R(3), RW(3), PM_frac(3)

  Grid = FFT_Grid(1:3, WARP_GRID)

  i = 0
  Do z=1, Grid(3)
     Do y=1, Grid(2)
        Do x=1, Grid(1)
           i = i + 1
           PM = (/x,y,z/)   !* Determine R-space value
           Call CalcRealSpace(Grid, PM, R)
           Call CalcWarpSpace(Warp, R, RW)  !** Warped R Value
           Call CalcGridSpace(Grid, RW, PM_Frac) !** Map it on the warped grid
           PM = PM_Frac  !** We don't care about the fracional coords
           Call WrapIndex(Grid, PM)  !* Wrap the mesh point into the unit cell
           
           !** Finally store the mapping
           Warp%Map(i) = CalcGridOffset(Grid, PM)
        End Do
     End do
  End do

  Return
End subroutine


!******************************************************************************
!
!  Warp_Apply_Mapping - Applies the mapping
!
!         Warp - Context to use
!         OldPsi - Old function in R-space
!         NewPsi - New function in R-space
!
!******************************************************************************

Subroutine Warp_Apply_Mapping(Warp, OldPsi, NewPsi)
  Type (Warp_context), Pointer :: Warp
  Complex,         Intent(IN)  :: OldPsi(:)
  Complex,         Intent(OUT) :: NewPsi(:)

  NewPsi = OldPsi(Warp%Map)   

  Return
End subroutine


!******************************************************************************
! 
!  Warp_AfterMove - Rotuine to call AFTER the atom's have been moved.
!        This routine calculates the transformation and applies it 
!        to all the wave functions and then destroys the Warp context
!
!        Warp - Context containing the Before move data
!
!******************************************************************************

Subroutine Warp_AfterMove(Warp)
  Type (Warp_context), Pointer :: Warp

  Integer :: i, ierr, Kpnt, Psi_toUse(Mem_MapSize)
  Complex, Pointer :: PsiR(:), NewPsi(:)
  Type (Mem_handle), Pointer :: Psi

  Call Warp_Store_Pos(Warp, NEW_POS)  !** Store the new positions
  Call Warp_Create_transform(Warp)    !** Create the coordinate transformation
  Call Warp_Create_Mapping(warp)      !** Create the grid mapping

  !** Now we're ready to apply the mapping **

  !** 1st create temp space
  Allocate(PsiR(FFT_Grid(4,WARP_GRID)), &  
       NewPsi(FFT_Grid(4,WARP_GRID)), STAT=ierr)
  Call Check_Error(ierr, "Could Not Allocate Temp arrays!", Error_Unit, &
       .TRUE.,paw_wc, "Warp_AfterMove:")
 
  Psi_toUse = MH_Skip    !** Now determine the set of Psi's to process **
  Where (PsiInfo(:)%Kpnt > 0) Psi_toUse = MH_toProcess
    
  !** Configure the psi's for loading.  Don't care about projectors so
  !** the kpnt is just set to 1.
  Call Phase_Generic(Psi_toUse, 1)  

  !** Start the looping
  Call GetNextPsi(Kpnt, Psi)
  Do While (Kpnt > 0)
     PsiR = 0          !** Convert to R-space
     Call V_G_to_R(Psi%Ptr, G_LOW, PsiR, WARP_GRID)

     Call Warp_Apply_Map(Warp, PsiR, NewPsi) !** Apply the transformation
    
     Psi%Ptr = 0     !** Switch back to G-space
     Call V_R_to_G(NewPsi, WARP_GRID, Psi%Ptr, G_LOW)

     Call SavePsi_tobuffer(Psi%Index, Psi%Ptr) !** Mark the Psi as changed
  End Do

  DeAllocate(PsiR, NewPsi)  !** Free the temp space


  Call Destroy_warp(warp)  !** Finally destry the context

  Return
End Subroutine

End Module

